/* ====================================================================
 *  Copyright 2004 Edward Rudd and Paul Querna
 *
 *  Licensed under the Apache License, Version 2.0 (the "License");
 *  you may not use this file except in compliance with the License.
 *  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 *  Some code based off of Subversion:
 *     Copyright (c) CollabNet.  All rights reserved.
 *
 */
 
#include "apr_errno.h"
#include "apr_general.h"
#include "apr_getopt.h"
#include "apr_strings.h"
#include "apr_pools.h"
#include "apr_dbm.h"
#include "svn_repos.h"
#include "svn_opt.h"
#include "svn_path.h"

#include "mod_svn_view_config.h"

#define SVN_VIEW_VERSION VERSION

#define PERROR(err)  if (err!=SVN_NO_ERROR) { \
        printf("Error was: %s\n",err->message); \
 }
 
static const apr_getopt_option_t gArgOptions[] = {
   // long name, short name, argc, help string
   {"help", 'h', 0, "Show Help"},
   {"debug", 'd', 0, "Show detailed debuging information"},
   {"create", 'c', 0, "Create a new svn_view user cache"},
   {"path", 'p', 1, "Path to the User Cache"},
   {"add", 'a', 1, "Add a new Repository to an existing user cache"},
   {"update", 'u', 1, "Update an existing user cache"},
   {0, 0, 0, 0}
};


typedef enum sv_cache_action sv_cache_action;
enum sv_cache_action {
    SVCA_NONE,
    SVCA_CREATE,
    SVCA_ADD,
    SVCA_UPDATE,
};

typedef struct sv_cache_rec sv_cache_rec;
struct sv_cache_rec
{
    apr_pool_t *p;
    int debug;
    sv_cache_action action;
    const char *svc_path;       /* Path to our cache directory */
    const char *core_path;       /* Path to our core dbm file */
    
    const char *repo_path;       /* Selected repository filesystem path */
    svn_revnum_t request_rev;   /* Selected revision */
    svn_repos_t *repo;          /* Repository handle */
    svn_fs_t *repo_fs;          /* Repository file system handle */
    svn_fs_root_t *rev_root;    /* file system revision root */
};

static apr_status_t svc_create(sv_cache_rec* pSvc);

static void sv_debug(sv_cache_rec* pSvc, const char* pMsg) {
    if(pSvc->debug == 1) {
        fprintf(stderr,"[debug] %s\n", pMsg);
    }
}

static void show_usage(apr_pool_t *pPool) 
{
    apr_size_t i;
    fprintf(stdout,"Usage: sv_ucache [options]\n\n");
    fprintf(stdout,"Valid options:\n");
    for (i = 0; gArgOptions[i].name && gArgOptions[i].optch; i++) {
        const char *optstr;
        svn_opt_format_option(&optstr, gArgOptions + i, TRUE, pPool);
        fprintf(stdout, "  %s\n", optstr);
    }
    fprintf(stdout, "\n");
}

static void show_help(apr_pool_t *pPool, int pType) 
{
    if(pType == 0) {
        fprintf(stdout,"sv_ucache - manage dbm cache for optional mod_svn_view features\n\n");
        fprintf(stdout,"\tPart of mod_svn_view %s\n",SVN_VIEW_VERSION);
        fprintf(stdout,"\tCopyright 2004 Edward Rudd and Paul Querna\n\n");
    }

    show_usage(pPool);

    if(pType == 0) {
        fprintf(stdout,"Common Operations:\n");
        fprintf(stdout,"\tCreate new Cache:\n");    
        fprintf(stdout,"\t\tsv_ucache --create --path /home/svn/svn_view_cache\n\n");
        fprintf(stdout,"\tAdd Repository:\n");
        fprintf(stdout,"\t\tsv_ucache --add /var/svn/mod_svn_view --path /home/svn/svn_view_cache\n\n");
        fprintf(stdout,"\tFull Update:\n");
        fprintf(stdout,"\t\tsv_ucache --update --path /home/svn/svn_view_cache\n\n");
    }
}

int main(int pArgc, const char *pArgv[])
{
    apr_pool_t *tPool;
    apr_getopt_t *tOptions;
    int tOptName;
    const char *tOptArg;
    sv_cache_rec *tSvc;
    apr_status_t tStatus;


    setbuf(stdout, NULL);

    if (apr_initialize() != APR_SUCCESS) {
        fprintf(stderr, "Unable to Initialize APR!\n");
        return -1;
    }
    if (apr_pool_create(&tPool, NULL) != APR_SUCCESS) {
        fprintf(stderr, "Unable to create Root Pool!\n");
        return -1;
    }

    if(pArgc <= 1) {
        show_help(tPool, 0);
        return -1;
    }

    tSvc = apr_pcalloc(tPool, sizeof(*tSvc));
    tSvc->action = SVCA_NONE;
    tSvc->debug = 0;
    tSvc->p = tPool;
    tSvc->repo_path = NULL;
    tSvc->svc_path = NULL;

    apr_getopt_init(&tOptions, tPool, pArgc, pArgv);

    while (1) {
        
        tStatus = apr_getopt_long(tOptions, gArgOptions, &tOptName, &tOptArg);
        if (APR_STATUS_IS_EOF(tStatus)) {
            break;
        }
        else if (tStatus != APR_SUCCESS) {
            fprintf(stderr, "Error: Invalid Arguments\n\n");
            show_help(tSvc->p, 1);
            return -1;
        }

        switch(tOptName) 
        {
            case 'h':
                show_help(tSvc->p, 0);
                return -1;
            case 'd':
                tSvc->debug = 1;
                break;
            case 'c':
                tSvc->action = SVCA_CREATE;
                break;
            case 'a':
                tSvc->action = SVCA_ADD;
                break;
            case 'u':
                tSvc->action = SVCA_UPDATE;
                break;
            case 'p':
                svn_path_get_absolute(&tSvc->repo_path, tOptArg, tSvc->p);
                break;
            case 'f':
                svn_path_get_absolute(&tSvc->svc_path, tOptArg, tSvc->p);
                break;
        }      
    }

    if(SVCA_NONE == tSvc->action) {
        fprintf(stderr, "Error: No Command Called\n\n");
        show_help(tSvc->p, 1);        
        return -1;
    }

    if(!tSvc->svc_path) {
        fprintf(stderr, "Error: This Command requires --cache\n\n");
        show_help(tSvc->p, 1);        
        return -1;
    }

    if(!tSvc->repo_path) {
        fprintf(stderr, "Error: This command Requires --path\n");
        show_help(tSvc->p, 1);
        return -1;
    }

    if(SVCA_CREATE == tSvc->action) {
        sv_debug(tSvc, "Starting Create Action");
        svc_create(tSvc);
        sv_debug(tSvc, "Done With Create");
    }
    else if(SVCA_ADD == tSvc->action) {

    }
    else if(SVCA_UPDATE == tSvc->action) {

    }

    return 0;    
}

static apr_status_t svc_create(sv_cache_rec* pSvc) 
{
    apr_finfo_t tFinfo;
    apr_status_t tStatus;
    apr_fileperms_t tDefaultPerms = APR_OS_DEFAULT;
    apr_dbm_t *tCoreDBM;
    apr_datum_t tKey, tVal;
    
    // Step One: Check for existing files/folders
    sv_debug(pSvc, "calling apr_start()");
    tStatus = apr_stat(&tFinfo, pSvc->svc_path, APR_FINFO_MIN, pSvc->p);
    if (APR_STATUS_IS_ENOENT(tStatus)) { 
        sv_debug(pSvc, "Got APR_NOFILE");
        tFinfo.filetype = APR_NOFILE;
    }

    if (APR_NOFILE != tFinfo.filetype) {
            sv_debug(pSvc, "Did not get APR_NOFILE");
            fprintf(stderr, "Error: Path Already Exists!\n\n");
            exit(-1);
    }

    // Step Two: Make our directory
    sv_debug(pSvc, "Creating Directories");
    tStatus = apr_dir_make_recursive (pSvc->svc_path, tDefaultPerms, pSvc->p);

    if (tStatus != APR_SUCCESS) {
            fprintf(stderr, "Error: Could not create path: %s\n\n", pSvc->svc_path);
            exit(-1);    
    }

    sv_debug(pSvc, "Joining Path");
    // TODO: This is not win32 safe
    pSvc->core_path =  svn_path_join(pSvc->svc_path, "core",pSvc->p);
    
    // Step Three: Get User Input

    // Step Four: Create Databases

    // Step Five: Insert Initial Data

    // Step Six: Review Setup
    return APR_SUCCESS;
}
